#include <stdio.h>
#include "StrProtocol.h"

Step::Step( unsigned char* inVals, int inLen, unsigned char* outVals, int outLen ) {
	inValues = inVals;
	inLength = inLen;
	outValues = outVals;
	outLength = outLen;
	currentInOffset = 0;
	currentOutOffset = 0;
}

void Step::startUp() {
	currentInOffset = 0;
	currentOutOffset = 0;
}

int Step::getNextInValue() {
	if ( currentInOffset >= inLength ) {
		return( -1 );
	} else {
		int inOffset = currentInOffset;
		currentInOffset++;
		return( *( inValues + inOffset ));
	}
}

int Step::getNextOutValue() {
	//printf( "..getNextOutValue, currentOffset is: %d, outLength is: %d\n", currentOutOffset, outLength );
	if ( currentOutOffset >= outLength ) {
		return( -1 );
	} else {
		int outOffset = currentOutOffset;
		currentOutOffset++;
		return( *( outValues + outOffset ));
	}
}

bool Step::checkReceivedInt( unsigned char val ) {
	int ckVal = getNextInValue();
	if ( ckVal == -1 ) {
		return( false );
	} else {
		unsigned char ck = (unsigned char)ckVal;
		return( ck == val );
	}
}

StrProtocol::StrProtocol() {
	init();
}

void StrProtocol::init() {
	for ( int i = 0; i < STEPCOUNT; i++ ) {
		steps[i] = NULL;
	}
	currentStep = 0;
	lastStep = 0;
}

bool StrProtocol::hasMoreSteps() {
	return( currentStep < lastStep );
}

void StrProtocol::setString( char* str ) {
	unsigned char inBuf[100];
	int inIdx;
	unsigned char outBuf[100];
	int outIdx;
	int val = 0;
	bool valExists = false;
	bool valIsIn = false;

	inIdx = 0;
	outIdx = 0;
	while ( *str ) {
		//printf( "char %c..", *str );
		if (( *str == '<' ) || ( *str == '>' )) {
			if (( inIdx > 0 ) && ( outIdx > 0 )) {
				addStep( &inBuf[0], inIdx, &outBuf[0], outIdx );
				inIdx = 0;
				outIdx = 0;
			}
		}
		if ( *str == '<' ) {
			valIsIn = true;
		} else if ( *str == '>' ) {
			valIsIn = false;
		} else if ( *str == ' ' ) {
			if ( valExists ) {
				if ( valIsIn ) {
					inBuf[ inIdx++ ] = val;
					valExists = false;
					val = 0;
				} else {
					outBuf[ outIdx++ ] = val;
					valExists = false;
					val = 0;
				}
				//printf( "processed a value, inIdx %d, outIdx %d\n", inIdx, outIdx );
			}
		} else if (( *str >= '0' ) && ( *str <= '9' )) {
			val = val*16 + (*str - '0');
			valExists = true;
		} else if (( *str >= 'a' ) && ( *str <= 'f' )) {
			val = val*16 + (*str - 'a') + 10;
			valExists = true;
		} else if (( *str >= 'A' ) && ( *str <= 'F' )) {
			val = val*16 + (*str - 'A' ) + 10;
			valExists = true;
		}
		str++;
	}
	//printf( "done processing string\n" );
			if ( valExists ) {
				if ( valIsIn ) {
					inBuf[ inIdx++ ] = val;
					valExists = false;
					val = 0;
				} else {
					outBuf[ outIdx++ ] = val;
					valExists = false;
					val = 0;
				}
			}
	if (( inIdx > 0 ) && ( outIdx > 0 )) {
		addStep( &inBuf[0], inIdx, &outBuf[0], outIdx );
	}
}

#include <malloc.h>
#include <string.h>
#include <memory.h>

void StrProtocol::addStep( unsigned char* inVals, int inLen, unsigned char* outVals, int outLen ) {
	unsigned char* inValues = (unsigned char*)malloc( inLen );
	unsigned char* outValues = (unsigned char*)malloc( outLen );
	memcpy( inValues, inVals, inLen );
	memcpy( outValues, outVals, outLen );
	//printf( "ADDING A STEP" );
	steps[ lastStep++ ] = new Step( inValues, inLen, outValues, outLen );
}

void StrProtocol::loadProtocol( char* fileName ) {
	if (( fileName != NULL ) && ( strlen( fileName ) > 0 )) {
		FILE* fp = fopen( fileName, "r" );
		if ( fp != NULL ) {
			char line[300];
			while (fgets( &line[0], 300, fp ) != NULL ) {
				setString( &line[0] );
			}
			fclose( fp );
		}
	}
}
void StrProtocol::startUp() {
	currentStep = 0;
}

int StrProtocol::getNextIntToReceive() {
	if ( currentStep < lastStep ) {
		int result = steps[ currentStep ]->getNextInValue();
		if ( result == -1 ) {
			currentStep++;
		}
		return( result );
	} else {
		return( -1 );
	}
}

int StrProtocol::getNextIntToSend() {
	//printf( "getNextIntToSend...currentStep %d, lastStep %d..", currentStep, lastStep );
	if ( currentStep < lastStep ) {
		return( steps[ currentStep ]->getNextOutValue() );
	} else {
		return( -1 );
	}
}

int StrProtocol::getStepNumber() {
	return( currentStep );
}

bool StrProtocol::checkReceivedInt( int val ) {
	if ( currentStep < lastStep ) {
		return( steps[ currentStep ]->checkReceivedInt( (unsigned char)val ));
	} else {
		return( false );
	}
}

